import threading 
import time
import os
import sys
from . import part, multipart_collector, cookie, session, grpc_collector, ws_executor
from . import wsgi_executor, xmlrpc_executor, grpc_executor
from aquests.lib import producers, importer
from aquests.protocols.grpc import discover
from aquests.protocols.http import http_util
from skitai import was as the_was
from hashlib import md5
import random
import base64
from . import cookie
from .config import Config
from jinja2 import Environment, PackageLoader
from chameleon import PageTemplateLoader
try:
	import xmlrpc.client as xmlrpclib
except ImportError:
	import xmlrpclib

class AuthorizedUser:
	def __init__ (self, user, realm, info = None):
		self.name = user
		self.realm = realm
		self.info = info
		
				
class Saddle (part.Part):
	use_reloader = False
	debug = False
	
	# Session
	securekey = None
	session_timeout = None
	
	#WWW-Authenticate	
	access_control_allow_origin = None
	access_control_max_age = 0
	authenticate = False
	authorization = "digest"	
	realm = None
	users = {}	
	opaque = None	
	
	PRESERVES_ON_RELOAD = ["reloadables"]
	
	def __init__ (self, app_name):
		part.Part.__init__ (self)
		self.app_name = app_name
		self.home = None
		self.jinja_env = Environment (loader = PackageLoader (app_name)) or None
		self.chameleon = None
		self.lock = threading.RLock ()
		
		self.reloadables = {}
		self.last_reloaded = time.time ()
		self.cached_paths = {}
		self.handlers = {}		
		self.cached_rules = []
		self.config = Config (preset = True)
	
	def skito_jinja (self):
		self.jinja_overlay ("${", "}", "<%", "%>", "<!---", "--->")

	def jinja_overlay (
		self, 
		variable_start_string = "{{",
		variable_end_string = "}}", 
		block_start_string = "{%", 
		block_end_string = "%}", 
		comment_start_string = "{#",
		comment_end_string = "#}",
		line_statement_prefix = "%", 
		line_comment_prefix = "%%",
		**karg
	):
		from .patches import jinjapatch
		self.jinja_env = jinjapatch.overlay (self.app_name, variable_start_string, variable_end_string, block_start_string, block_end_string, comment_start_string, comment_end_string, line_statement_prefix, line_comment_prefix, **karg)
	
	def get_error_page (self, error):
		handler = self.handlers.get (error ['code'])
		if not handler:
			return
		return handler [0](the_was._get (), error)
		
	def add_error_handler (self, errcode, f, **k):
		self.handlers [errcode] = (f, k)		
		
	def errorhandler (self, errcode, **k):
		def decorator(f):
			self.add_error_handler (errcode, f, **k)
		return decorator
		
	def watch (self, module):
		self.reloadables [module] = self.get_file_info (module)
	
	def maybe_reload (self):
		if time.time () - self.last_reloaded < 1.0:
			return
			
		for module in list (self.reloadables.keys ()):			
			try:
				fi = self.get_file_info (module)
			except FileNotFoundError:
				del self.reloadables [module]
				continue
				
			if self.reloadables [module] != fi:				
				importer.reloader (module)
				self.reloadables [module] = fi
		
		self.last_reloaded = time.time ()
		
	def get_file_info (self, module):		
		stat = os.stat (module.__file__)
		return stat.st_mtime, stat.st_size
	
	def get_resource (self, *args):
		return os.path.join (self.home, "resources", *args)
	
	PACKAGE_DIRS = ["package", "appack", "contrib"]
	def add_package (self, *names):
		for name in names:
			self.PACKAGE_DIRS.append (name)
		
	def set_home (self, path):
		self.home = path
		self.chameleon = PageTemplateLoader (
			os.path.join(path, "templates"),
			auto_reload = self.use_reloader,
			restricted_namespace = False
		)
		
		package_dirs = []
		for d in self.PACKAGE_DIRS:
			maybe_dir = os.path.join (path, d)
			if os.path.isdir (maybe_dir):
				package_dirs.append (maybe_dir)
		
		for k, v in sys.modules.items ():
			try:
				modpath = v.__spec__.origin
			except AttributeError:
				continue
			
			if modpath:
				for package_dir in package_dirs:
					if modpath.startswith (package_dir):						
						self.watch (v)
						break
			
	def render (self, was, template_file, _do_not_use_this_variable_name_ = {}, **karg):
		while template_file and template_file [0] == "/":
			template_file = template_file [1:]	

		if _do_not_use_this_variable_name_: 
			assert not karg, "Can't Use Dictionary and Keyword Args Both"
			karg = _do_not_use_this_variable_name_

		karg ["was"] = was		
		template = self.get_template (template_file)
		self.when_got_template (was, template, karg)
			
		rendered = template.render (**karg)
		self.when_template_rendered (was, template, karg, rendered)
		return rendered
					
	def get_template (self, name):
		if name.endswith ('.pt') or name [-5:] == ".ptal":
			return self.chameleon [name]
		return self.jinja_env.get_template (name)		
			
	def get_www_authenticate (self):
		if self.authorization == "basic":
			return 'Basic realm="%s"' % self.realm
		else:	
			if self.opaque is None:
				self.opaque = md5 (self.realm.encode ("utf8")).hexdigest ()
			return 'Digest realm="%s", qop="auth", nonce="%s", opaque="%s"' % (
				self.realm, http_util.md5uniqid (), self.opaque
			)
	
	def get_password (self, user):
		info = self.users.get (user)
		if not info:
			return None, 0 # passwrod, encrypted
		return type (info) is str and (info, 0) or info [:2]
	
	def get_info (self, user):
		info = self.users.get (user)
		if not info: return None
		return (type (info) is not str and len (info) == 3) and info [-1] or None
				
	def authorize (self, auth, method, uri):
		if self.realm is None or not self.users:
			return
				
		if auth is None:
			return self.get_www_authenticate ()
		
		# check validate: https://evertpot.com/223/
		amethod, authinfo = auth.split (" ", 1)
		if amethod.lower () != self.authorization:
			return self.get_www_authenticate ()
		
		if self.authorization == "basic":
			basic = base64.decodestring (authinfo.encode ("utf8")).decode ("utf8")
			current_user, current_password = basic.split (":", 1)
			password, encrypted = self.get_password (current_user)
			if encrypted:
				raise AssertionError ("Basic authorization can't handle encrypted password")
			if password ==  current_password:
				return AuthorizedUser (current_user, self.realm, self.get_info (current_user))
				
		else:
			method = method.upper ()
			infod = {}
			for info in authinfo.split (","):
				k, v = info.strip ().split ("=", 1)
				if not v: return self.get_www_authenticate ()
				if v[0] == '"': v = v [1:-1]
				infod [k]	 = v
			
			current_user = infod.get ("username")
			if not current_user:
				return self.get_www_authenticate ()
			
			password, encrypted = self.get_password (current_user)
			if not password:
				return self.get_www_authenticate ()
				
			try:
				if uri != infod ["uri"]:					
					return self.get_www_authenticate ()
				if encrypted:	
					A1 = password
				else:
					A1 = md5 (("%s:%s:%s" % (infod ["username"], self.realm, password)).encode ("utf8")).hexdigest ()
				A2 = md5 (("%s:%s" % (method, infod ["uri"])).encode ("utf8")).hexdigest ()
				Hash = md5 (("%s:%s:%s:%s:%s:%s" % (
					A1, 
					infod ["nonce"],
					infod ["nc"],
					infod ["cnonce"],
					infod ["qop"],
					A2
					)).encode ("utf8")
				).hexdigest ()

				if Hash == infod ["response"]:
					return AuthorizedUser (current_user, self.realm, self.get_info (current_user))
					
			except KeyError:
				pass
		
		return self.get_www_authenticate ()
	
	def is_allowed_origin (self, request, allowed_origins):
		origin = request.get_header ('Origin')
		if not origin:
			return True
		if not allowed_origins:
			allowed_origins = ["%s://%s" % (request.get_scheme (), request.get_header ("host", ""))]
		elif "*" in allowed_origins:
			return True		
		if origin in allowed_origins:
			return True
		return False
		
	def is_authorized (self, request, authenticate):
		if not authenticate:
			return True
		
		www_authenticate = self.authorize (request.get_header ("Authorization"), request.command, request.uri)
		if type (www_authenticate) is str:
			request.response.set_header ('WWW-Authenticate', www_authenticate)
			return False
		elif www_authenticate:
			request.user = www_authenticate				
		return True
	
	def set_devel (self, debug = True, use_reloader = True):
		self.debug = debug
		self.use_reloader = use_reloader
	
	def get_collector (self, request):
		ct = request.get_header ("content-type")
		if not ct: return
		if ct.startswith ("multipart/form-data"):
			return multipart_collector.MultipartCollector
			
		if ct.startswith ("application/grpc"):
			try:
				i, o = discover.find_type (request.uri [1:])
			except KeyError:
				raise NotImplementedError			
			return grpc_collector.grpc_collector			
					
	def get_method (self, path_info, request):		
		command = request.command.upper ()
		content_type = request.get_header_noparam ('content-type')
		authorization = request.get_header ('authorization')		
		current_app, method = self, None
		
		with self.lock:
			if self.use_reloader:
				self.maybe_reload ()
				
			try:
				method, options = self.cached_paths [path_info]
				kargs = {}
			except KeyError:
				ind = 0				
				for rulepack, _method, _options in self.cached_rules:
					_found, _kargs = self.try_rule (path_info, rulepack [0], rulepack [1])
					if _found:
						method = _method
						kargs = _kargs
						options = _options
						break
					ind += 1
									
		if not method:
			if self.use_reloader:
				self.lock.acquire ()
			try:	
				current_app, method, kargs, options, match, matchtype = self.get_package_method (path_info, command, content_type, authorization, self.use_reloader)
			finally:	
				if self.use_reloader: 
					self.lock.release ()
			
			if matchtype == -1:
				return current_app, method, None, options, 301
			
			if not method:
				return current_app, None, None, options, 404
				
			if not self.use_reloader:
				if matchtype == 1:
					with self.lock:
						self.cached_paths [match] = (method, options)
								
				elif matchtype == 2:
					with self.lock:
						self.cached_rules.append ([match, method, options])
						# sort by length of rule desc
						self.cached_rules.sort (key = lambda x: len (x [0][-1][-2]), reverse = True)
		
		resp_code = 0
		if options:
			allowed_types = options.get ("content_types", [])
			if allowed_types and content_type not in allowed_types:
				return current_app, None, None, options, 415 # unsupported media type
			
			allowed_methods = options.get ("methods", [])			
			if allowed_methods and command not in allowed_methods:				
				return current_app, None, None, options, 405 # method not allowed
			
			if command == "OPTIONS":				
				request_method = request.get_header ("Access-Control-Request-Method")
				if request_method and request_method not in allowed_methods:
					return current_app, None, None, options, 405 # method not allowed
			
				response = request.response
				response.set_header ("Access-Control-Allow-Methods", ", ".join (allowed_methods))
				access_control_max_age = options.get ("access_control_max_age", self.access_control_max_age)	
				if access_control_max_age:
					response.set_header ("Access-Control-Max-Age", str (access_control_max_age))
				
				requeste_headers = request.get_header ("Access-Control-Request-Headers", "")		
				if requeste_headers:
					response.set_header ("Access-Control-Allow-Headers", requeste_headers)
					
				resp_code = 200
			
			else:
				if not self.is_allowed_origin (request, options.get ("access_control_allow_origin", self.access_control_allow_origin)):
					resp_code =  403
				elif not self.is_authorized (request, options.get ("authenticate", self.authenticate)):
					resp_code =  401				
		
		if resp_code in (401, 200):
			authenticate = options.get ("authenticate", self.authenticate)
			if authenticate:
				request.response.set_header ("Access-Control-Allow-Credentials", "true")
							
		access_control_allow_origin = options.get ("access_control_allow_origin", self.access_control_allow_origin)
		if access_control_allow_origin and access_control_allow_origin != 'same':
			request.response.set_header ("Access-Control-Allow-Origin", ", ".join (access_control_allow_origin))
		
		return current_app, method, kargs, options, resp_code
	
	def restart (self, wasc, route):
		self.start (wasc, route, self.packages)
			
	def create_on_demand (self, was, name):
		class G: 
			pass
		
		# create just in time objects
		if name == "cookie":
			return cookie.Cookie (was.request, self.securekey, self.route [:-1], self.session_timeout)
			
		elif name in ("session", "mbox"):
			if not was.in__dict__ ("cookie"):
				was.cookie = cookie.Cookie (was.request, self.securekey, self.route [:-1], self.session_timeout)			
			if name == "session":
				return was.cookie.get_session ()
			if name == "mbox":
				return was.cookie.get_notices ()
				
		elif name == "g":
			return G ()
		
	def cleanup_on_demands (self, was):
		if was.in__dict__ ("g"):
			del was.g
		if not was.in__dict__ ("cookie"):
			return
		for j in ("session", "mbox"):
			if was.in__dict__ (j):		
				delattr (was, j)
		del was.cookie
										
	def __call__ (self, env, start_response):
		was = env ["skitai.was"]		
		was.app = self
		was.ab = self.build_url
		was.response = was.request.response
		
		content_type = env.get ("CONTENT_TYPE", "")				
		if content_type.startswith ("text/xml") or content_type.startswith ("application/xml"):
			result = xmlrpc_executor.Executor (env, self.get_method) ()
		elif content_type.startswith ("application/grpc"):
			result = grpc_executor.Executor (env, self.get_method) ()			
		elif env.get ("websocket.params"):
			result = ws_executor.Executor (env, None) ()
		else:	
			result = wsgi_executor.Executor (env, self.get_method) ()
		
		del was.response
		del was.ab		
		self.cleanup_on_demands (was) # del session, mbox, cookie, g
			
		return result
		


import sys
import subprocess
import os		
import signal
import time
from aquests.lib import confparse, logger, flock, pathtool
from skitai.server.wastuff import process, daemon
import time
			
class Service (daemon.Daemon):
	BACKOFF_MAX_INTERVAL = 600
	CLEAN_SHUTDOWNED = {}
	RESTART_QUEUE = {}
	DAEMONS = ("smtpda", "cron")
	
	def __init__ (self, cmd, logpath, varpath, verbose):
		self.cmd = cmd
		self.logpath = logpath
		self.varpath = varpath
		self.consol = verbose
		self.make_logger (False)

		self.backoff_start_time = None
		self.backoff_interval = 5
		self.child = None
		self.loop = 1
	
	def set_backoff (self, reset = False):
		if reset:
			if self.backoff_start_time is None: 
				return
			else:		
				self.backoff_start_time = None
				self.backoff_interval = 5
				return
		
		if self.backoff_start_time is None:
			self.backoff_start_time = time.time ()
	
	def shutdown (self):
		self.logger ("[info] try to kill %s..." % self.child.name)
		self.child.kill ()
		for i in range (30):
			time.sleep (1)
			if self.child.poll () is None:
				self.logger ("[info] %s is still alive" % self.child.name)				
			else:
				break
		
		if self.child.poll () is None:
			self.logger ("[info] force to kill %s" % self.child.name)
			self.child.send_signal ('kill')
	
	def hTERM (self, signum, frame):			
		self.loop = 0
		
	def run (self):
		if os.name == "nt":
			signal.signal(signal.SIGBREAK, self.hTERM)
		else:
			signal.signal(signal.SIGTERM, self.hTERM)
	
		try:
			try:
				self.start ()
			except KeyboardInterrupt:
				pass
			except:
				self.logger.trace ()
		finally:
			self.shutdown ()		
	
	def create (self):		
		self.child = process.Process (
			self.cmd, 
			'instance',
			self.varpath and self.varpath or daemon.get_default_varpath ()
		)
	
	def start (self):
		self.create ()
		try:
			while self.loop:
				exitcode = self.child.poll ()			
				if exitcode is None:
					self.set_backoff (True)
					continue
				
				if exitcode == 0:
					self.logger ("[info] instance has been shutdowned cleanly")
					break
					
				elif exitcode == 3:
					self.logger ("[info] try re-starting up instance")
					self.create ()
					
				else:
					self.set_backoff ()
					if time.time() - self.backoff_start_time >= self.backoff_interval:
						self.logger ("[fail] instance encountered unexpected error and terminated, try re-starting up (current backoff interval is %d)" % self.backoff_interval)
						self.backoff_interval = self.backoff_interval * 2
						if self.backoff_interval > self.BACKOFF_MAX_INTERVAL:
							self.backoff_interval = self.BACKOFF_MAX_INTERVAL
						self.create ()
				time.sleep (1)
		
		except KeyboardInterrupt:
			pass	

	
if __name__ == "__main__":
	service = Service ()	
	service.run ()
	


# Patch for using with Vue.js namespace and v-on shortcut
# 2017. 2.17 by Hans Roh
# 2017. 2.22 This patch needn't from Chameleon version 3.1

from chameleon import tokenize, parser
import re

a = tokenize.collector.add

# change only this line, just add @ for v-on shortcut
a ("NameStrt", "[A-Za-z_:@]|[^\\x00-\\x7F]")

# these're just consequence for recompiling tokenize.re_xml_spe
a ("Name", "(?:%(NameStrt)s)(?:%(NameChar)s)*")
a("ElemTagCE",
  "(%(Name)s)(?:(%(S)s)(%(Name)s)(((?:%(S)s)?=(?:%(S)s)?)"
  "(?:%(AttValSE)s|%(Simple)s)|(?!(?:%(S)s)?=)))*(?:%(S)s)?(/?>)?")
a("MarkupSPE",
  "<(?:!(?:%(DeclCE)s)?|"
  "\\?(?:%(PI_CE)s)?|/(?:%(EndTagCE)s)?|(?:%(ElemTagCE)s)?)")
a("XML_SPE", "%(TextSE)s|%(MarkupSPE)s")

tokenize.re_xml_spe = re.compile(tokenize.collector.res['XML_SPE'])

# for ignoring namespace exception
def unpack_attributes(attributes, namespace, default):
    namespaced = parser.OrderedDict()

    for index, attribute in enumerate(attributes):
        name = attribute['name']
        value = attribute['value']
        if ':' in name:
            prefix = name.split(':')[0]
            name = name[len(prefix) + 1:]
            try:
                ns = namespace[prefix]
            except KeyError:
                # exception disabled by Hans
                ns = default
                #raise KeyError(
                #    "Undefined namespace prefix: %s." % prefix)
        else:
            ns = default
        namespaced[ns, name] = value

    return namespaced
    
parser.unpack_attributes  = unpack_attributes



"""
2015. 12. 10
Hans Roh
"""
from . import Saddle, part

Saddle = Saddle.Saddle
Saddlery = part.Part

request_started = 0
request_finished = 1
request_exception_occured = 2
request_tearing_down = 3
template_rendered = 4
message_flashed = 5




import os
import subprocess, os, sys, signal, subprocess
from aquests.lib import flock

class Process:
	def __init__ (self, cmd, name, vardir = None):
		self.cmd = cmd
		self.name = name
		self.vardir = vardir		
		self.child = None
		if os.name == "nt":
			self.flock = flock.Lock (os.path.join (self.vardir, ".%s" % self.name))			
		self.create ()
	
	def create (self):	
		if os.name == "nt":
			self.child = subprocess.Popen (
				self.cmd, 
				shell = True,
				creationflags = subprocess.CREATE_NEW_PROCESS_GROUP
			)							
		else:
			self.child = subprocess.Popen (
				"exec " + self.cmd, 
				shell = True
			)

	def send_signal (self, req):
		if req == "start": 
			req = "restart"
		else:
			if req not in ("terminate", "kill", "restart", "rotate"):
				print("[error] unknown command")
				sys.exit (1)
		
		pid	 = self.child.pid
		if os.name == "nt":		
			self.flock.lock ("signal", req)
			
			if req == "terminate":
				sig = signal.CTRL_C_EVENT
			elif req == "kill":	
				sig = signal.CTRL_BREAK_EVENT
			
			try:
				os.kill (pid, sig)
			except ProcessLookupError:
				pass	
					
		else:			
			if pid:					
				if req == "terminate": sig = signal.SIGTERM				
				elif req == "kill": sig = signal.SIGKILL
				elif req == "restart": sig = signal.SIGHUP				
				elif req == "rotate": sig = signal.SIGUSR1
				try:
					os.kill (pid, sig)
				except ProcessLookupError:
					pass
						
			elif req == "restart":
				self.flock.lock ("signal", req)
			
	def kill (self):
		self.send_signal ("terminate")
	
	def poll (self):
		return self.child.poll ()
		


import os
import sys

class Daemonizer:
	def __init__(self, chdir="/", umask=0o22):
		self.chdir = chdir
		self.umask = umask
		self.pidfile = os.path.join (chdir, '.pid')
		
	def runAsDaemon(self):
		self.fork_and_die()
		self.dettach_env ()		
		self.fork_and_die()	
		
		sys.stdout.flush()
		sys.stderr.flush()
		self.attach_stream('stdin', 'r')
		self.attach_stream('stdout', 'a+')
		self.attach_stream('stderr', 'a+')
		
	def dettach_env (self):
		os.setsid()
		os.umask(self.umask)
		os.chdir(self.chdir)
	
	def attach_stream (self, name, mode, fd = '/dev/null'):
		stream = open(fd, mode)
		os.dup2(stream.fileno(), getattr(sys, name).fileno())
		
	def fork_and_die(self):
		r = os.fork()
		if r == -1:
			raise OSError("Couldn't fork().")
		elif r > 0:  # I'm the parent
			if self.pidfile: 
				open (self.pidfile, 'w').write (str(r))
			sys.exit(0)
		elif r < 0:
			raise OSError("Something bizarre happened while trying to fork().")
			# now only r = 0 (the child) survives.
		return r

if __name__ == "__main__"	:
	import time
	
	Daemonizer ().runAsDaemon ()	
	f = open ('/home/ubuntu/out', 'w')
	while 1:
		time.sleep (1)
		f.write ('asdkljaldjalkdjalkdsa\n')
		f.flush()
	f.close ()

		



import re
import stat
import time
from . import mime_type_table
import os
from . import filesys
from skitai.server import http_date
from aquests.lib import producers
from aquests.protocols.http.http_util import *
from hashlib import md5

IF_MODIFIED_SINCE = re.compile (
	'([^;]+)((; length=([0-9]+)$)|$)',
	re.IGNORECASE
	)

IF_NONE_MATCH = re.compile (
	'"?(.+?)(?:$|")',
	re.IGNORECASE
	)
	
USER_AGENT = re.compile ('User-Agent: (.*)', re.IGNORECASE)

CONTENT_TYPE = re.compile (
	r"Content-Type: ([^;]+)((; boundary=([A-Za-z0-9'\(\)+_,./:=?-]+)$)|$)",
	re.IGNORECASE
	)

def get_re_match (head_reg, value):
	m = head_reg.match (value)
	if m and m.end() == len (value):
		return m
	return ''		
	
class MemCache:
	def __init__ (self):
		self.d = {}
	
	def read (self, path, file, etag):	
		cached = self.d.get (path)
		if cached is None or cached [0] != etag:			
			with open (file, "rb") as f:
				self.d [path] = (etag, f.read ())
		return self.d [path][1]

			
class Handler:
	directory_defaults = [
		'index.htm',
		'index.html'
		]
		
	def __init__ (self, wasc, pathmap = {}, max_ages = None, alt_handlers = []):
		self.wasc = wasc
		
		self.memcache = MemCache ()
		self.max_ages = []
		if max_ages:
			self.max_ages = [(k, v) for k, v in max_ages.items ()]
			self.max_ages.sort (key = lambda x: len (x [0]), reverse = True)
			
		self.alt_handlers = alt_handlers
		self.permit_cache = {}
		self.filesystem = filesys.mapped_filesystem ()
		for k, v in list(pathmap.items ()):
			self.add_route (k, v)
	
	def close (self):	
		for h in self.alt_handlers:
			try: h.close ()
			except AttributeError: pass
										
	def match (self, request):
		return 1
	
	def add_route (self, alias, option):
		self.filesystem.add_map (alias, option)
	
	def handle_alternative (self, request):
		for h in self.alt_handlers:
			if h.match (request):
				try:
					h.handle_request (request)
				except:						
					self.wasc.logger.trace("server")
					try: request.response.error (500)
					except: pass
				return
		request.response.error (404)
		
	def isprohibited (self, request, path):
		dirname = os.path.split(path) [0]				
		permission = self.filesystem.get_permission (dirname)
		if not permission:
			return False
		if not self.wasc.authorizer.has_permission (request, permission):
			return True			
		return False	
		
	def handle_request (self, request):
		if request.command not in ('get', 'head'):
			self.handle_alternative (request)
			return
		
		path, params, query, fragment = request.split_uri()
		if '%' in path:
			path = unquote (path)
		
		# return redirect
		if path == "":
			request.response['Location'] = '/'
			request.response.error (301)
			return
		
		if path.find ("./") != -1 or path.find ("//") != -1:
			request.response.error (404)
			return
				
		while path and path [0] == '/':
			path = path[1:]
		
		if self.filesystem.isdir (path):		
			if path and path[-1] != '/':
				request.response['Location'] = '/%s/' % path
				request.response.error (301)
				return
			
			found = False
			for default in self.directory_defaults:
				p = path + default
				if self.filesystem.isfile (p):
					path = p
					found = True
					break
			
			if not found:
				self.handle_alternative (request)
				return
		
		elif not self.filesystem.isfile (path):			
			return self.handle_alternative (request)
		
		if path and path [-1] == "/":
			return request.response.error (404)
					
		if self.isprohibited (request, path):
			return
		
		try:
			current_stat = self.filesystem.stat (path)
			mtime = current_stat [stat.ST_MTIME]
			file_length = current_stat [stat.ST_SIZE]			
		except:
			self.handle_alternative (request)
			return
		
		etag = self.make_etag (file_length, mtime)	
		inm_h = request.get_header ("if-none-match")	
		if inm_h:
			inm = get_re_match (IF_NONE_MATCH, inm_h)			
			if inm and etag == inm.group (1):
				self.set_cache_control (request, path, mtime, etag)
				request.response.start (304)
				request.response.done()
				return
		
		ims_h = request.get_header ("if-modified-since")
		if ims_h:
			ims = get_re_match (IF_MODIFIED_SINCE, ims_h)
			length_match = 1
			if ims:
				length = ims.group (4)
				if length:
					try:
						length = int(length)
						if length != file_length:
							length_match = 0
					except:
						pass
						
			ims_date = 0
			if ims:
				ims_date = http_date.parse_http_date (ims.group (1))
	
			if length_match and ims_date:
				if mtime <= ims_date:
					self.set_cache_control (request, path, mtime, etag)
					request.response.start (304)
					request.response.done()
				return
		
		request.response ['Content-Length'] = file_length
		self.set_cache_control (request, path, mtime, etag)
		self.set_content_type (path, request)
		
		if request.command == 'get':
			# if head, don't send contents
			if file_length < 4096:
				request.response.push (
					self.memcache.read (path, self.filesystem.translate (path), etag)
				)
			else:					
				request.response.push (producers.file_producer (
					self.filesystem.open (path, 'rb'))
				)
		request.response.done()
	
	def set_cache_control (self, request, path, mtime, etag):
		max_age = self.max_ages and self.get_max_age (path) or 0
		if request.version == "1.0":			
			request.response ['Last-Modified'] = http_date.build_http_date (mtime)
			if max_age:
				request.response ['Expires'] = http_date.build_http_date (mtime + max_age)
		else:
			request.response ['Etag'] = '"' + etag + '"'
			if max_age:
				request.response ['Cache-Control'] = "max-age=%d" % max_age		
				
	def get_max_age (self, path):
		path = not path and "/" or "/" + path
		for prefix, value in self.max_ages:			
			if path.startswith (prefix):
				return value
		
	def make_etag (self, file_length, mtime):
		return md5 (("%d:%d" % (file_length, int (mtime))).encode ("utf8")).hexdigest()
	
	def set_content_type (self, path, request):
		ext = get_extension (path).lower()
		if ext in mime_type_table.content_type_map:
			request.response['Content-Type'] = mime_type_table.content_type_map[ext]
		else:
			request.response['Content-Type'] = 'text/plain'
			



from .secured_cookie_value import SecuredDictValue
import time

class Session (SecuredDictValue):
	default_session_timeout = 1200 # 20 min.	
	KEY = "SESS"	
	
	def __init__ (self, name, cookie, request, secret_key, session_timeout = 0):
		self.session_timeout = session_timeout and session_timeout or self.default_session_timeout
		SecuredDictValue.__init__ (self, name, cookie, request, secret_key)
			
	def validate (self):
		if not '_expires' in self.data:
			self.data = {}
			return
			
		if type (self.data ['_expires']) is tuple:
			expires, addr = self.data ['_expires']
			self.__source_verified = (addr == self.request.get_remote_addr ())				
		else:
			expires = self.data ['_expires']
			
		if time.time() > expires: # expired
			self.data = {}
			return
			
		if self.session_timeout + time.time () < expires: # too long, maybe fraud
			self.data = {}
			return
	
	def getv (self, k, v = None):
		if not self.__source_verified:			
			self.data = {}
			self.dirty = True
			return v
		return self.get (k, v)
				
	def recal_expires (self, expires):
		if expires is None:
			return self.session_timeout
		if expires == "now":
			return 0
		if expires == "never":
			raise ValueError("session must be specified expires seconds")
		return int (expires)
		
	def commit (self, expires = None):
		# always commit for extending/expiring expires
		if not self.dirty or self.data is None: return
		expires = self.recal_expires (expires)
		self ["_expires"] = (time.time () + expires, self.request.get_remote_addr ())
		if len (self.data) == 1: # only have _expires, expire now
			expires = 0
		self.set_cookie (expires)		
		
	def set_default_session_timeout (self, timeout):
		self.default_session_timeout = timeout




from aquests.protocols.http import response as http_response, buffers
from aquests.lib import compressors
import time


class ProxyResponse (http_response.Response):
	SIZE_LIMIT = 2**24
	
	def __init__ (self, request, header, accept_gzip, client_request, asyncon):				
		self.client_request = client_request 
		self.asyncon = asyncon
		self.accept_gzip = accept_gzip		
		self.request = request
		self.header_s = header		
		if header [:2] == "\r\n":
			header = header [2:]
		header = header.split ("\r\n")		
		self.response = header [0]
		self.header = header [1:]
		self._header_cache = {}
		self.flushed_time = 0
		self.version, self.code, self.msg = http_response.crack_response (self.response)
		self.p, self.u = None, None
		self.decompressor = None
		self.gzip_compressed = False	
		self.is_xmlrpc_return = False
		self.make_decompressor = False
		
		content_encoding = self.get_header ("Content-Encoding")			
		if content_encoding == "gzip":
			if self.accept_gzip:
				self.gzip_compressed = True
			else:	
				self.make_decompressor = True
						
		self.size = 0
		self.got_all_data = False
		self.set_max_age ()
	
	def body_expected (self):
		cl = self.get_header ("Content-Length")
		if cl == 0:
			self.got_all_data = True
			return False
		
		te = self.get_header ("Transfer-Encoding")
		if cl is None and te != "chunked":
			hv = self.version
			cn = self.get_header ("Connection")			
			if cn is None:
				if hv == "1.0": cn = "close"
				else: cn = "keep-alive"	
			else:
				cn = cn.lower ()			
			if cn == "keep-alive":				
				self.got_all_data = True
				return False
		
		if self.p is None:
			self.p, self.u = buffers.getfakeparser (buffers.list_buffer, cache = self.max_age)
			if self.make_decompressor:
				self.decompressor = compressors.GZipDecompressor ()
			
		return True
		
	def get_header_lines (self):
		return self.header
			
	def init_buffer (self):
		# do this job will be executed in body_expected ()
		pass
		
	def is_gzip_compressed (self):
		return self.gzip_compressed
	
	def close (self):
		# channel closed and called automatically by channel
		self.asyncon.handle_abort ()		
		self.done ()
		#self.asyncon.handle_close (710, "Channel Closed")
		
	def ready (self):
		#print ('====== READYU', len (self.u), self.got_all_data)
		return len (self.u) or self.got_all_data
		
	def more (self):
		self.flushed_time = time.time ()
		data = self.u.read ()		
		#print ('-----', data [:70])
		return data
	
	def collect_incoming_data (self, data):
		http_response.Response.collect_incoming_data (self, data)
		


def shutdown (root):
	import md5
	import os
	from aquests.lib  import confparse	
	from aquests.lib  import timeoutsocket
	import urllib.request, urllib.parse, urllib.error
	
	cf = confparse.ConfParse (os.path.join (root, "etc/server.conf"))	
	port = cf.getint ("server", "port")
	timeoutsocket.setDefaultSocketTimeout (3)
	try:
		urllib.request.urlopen ("http://127.0.0.1:%d/admin/maintern/shutdown" % port)
	except AttributeError:
		pass	
	


from . import default_handler, wsgi_handler, proxypass_handler, websocket_handler, api_access_handler
import skitai
from . import http2_handler		
from .. import wsgi_apps
import os

class VHost:
	def __init__ (self, wasc, clusters, cachefs, static_max_ages, enable_apigateway, apigateway_authenticate, apigateway_realm, apigateway_secret_key):
		self.wasc = wasc
		self.clusters = clusters
		self.cachefs = cachefs
		
		self.apps = wsgi_apps.ModuleManager(self.wasc, self)
		self.handlers = []
		
		self.proxypass_handler = proxypass_handler.Handler (self.wasc, clusters, cachefs)
		if enable_apigateway:
			self.access_handler = api_access_handler.Handler (self.wasc, apigateway_authenticate, apigateway_realm, self.proxypass_handler, apigateway_secret_key)
			alternative_handlers = [self.access_handler]
		else:
			self.access_handler = None
			alternative_handlers = [self.proxypass_handler]			
		alternative_handlers.append (websocket_handler.Handler (self.wasc, self.apps))
		alternative_handlers.append (wsgi_handler.Handler (self.wasc, self.apps))
		self.default_handler = default_handler.Handler (self.wasc, {}, static_max_ages, alternative_handlers)		
		self.handlers.append (http2_handler.Handler (self.wasc, self.default_handler))
		
	def close (self):	
		for h in self.handlers:
			try: h.close ()
			except AttributeError: pass		
		self.apps.cleanup ()
	
	def set_auth_handler (self, storage):
		if self.access_handler:
			self.access_handler.set_auth_handler (storage)
						
	def add_proxypass (self, route, cname):
		self.proxypass_handler.add_route (route, cname)
		
	def add_route (self, route, target):
		self.default_handler.add_route (route, target)		
		
	def add_module (self, route, path, module, pref):
		self.apps.add_module (route, path, module, pref)


class Handler:
	def __init__ (self, wasc, *args):
		self.wasc = wasc
		self.vhost_args = args
		self.sites = {}
		self.__cache = {}
	
	def close (self):	
		for v in self.sites.values ():
			if hasattr (v, "close"):
				v.close ()
				
	def match (self, request):
		return self.find (request.get_header ("host")) and 1 or 0
		
	def handle_request (self, request):
		vhost = self.find (request.get_header ("host"))
		for h in vhost.handlers:
			if h.match (request):
				h.handle_request (request)
				break
			
	def add_route (self, rule, routepair, config = None):
		reverse_proxing = False
		if rule.strip () in ("*", "default"):
			rule = None
		else:
			rule = tuple (rule.split ())
		
		if rule not in self.sites:				
			self.sites [rule] = VHost (self.wasc, *self.vhost_args)

		vhost = self.sites [rule]
		route, target = [x.strip () for x in routepair.split ("=", 1)]
		if target.startswith ("@"):
			if route [-1] == "/":
				route = route [:-1]
			vhost.add_proxypass (route, target [1:].strip ())
			reverse_proxing = True
		
		elif os.path.isdir (target):
			if route [-1] == "/":
				route = route [:-1]
			vhost.add_route (route, target)
			
		else:
			fullpath = os.path.split (target.strip())
			vhost.add_module (route, os.sep.join (fullpath[:-1]), fullpath [-1], config)
		return reverse_proxing
		
	def find (self, host):
		if host:
			host = host.split (":", 1)[0]
			vhost = self.__cache.get (host)
			if vhost:
				return vhost
				
			for rules in self.sites:
				if rules is None: continue
				if host in rules:
					vhost = self.sites [rules]
					self.__cache [host] = vhost
					return vhost
					
				for rule in rules:
					if rule [0] == "." and host.endswith (rule) or host == rule [1:]:
						vhost = self.sites [rules]
						self.__cache [host] = vhost
						return vhost
						
		return self.sites.get (None, None)
	
	


import re, sys
try:
	from urllib.parse import unquote_plus, quote_plus, urljoin
except ImportError:
	from urllib import unquote_plus, quote_plus
	from urlparse import urljoin
import os
from aquests.lib import importer, strutil
from types import FunctionType as function
import inspect

	
RX_RULE = re.compile ("(/<(.+?)>)")

class Part:
	def __init__ (self, *args, **kargs):			
		self.module = None
		self.packagename = None
		self.wasc = None		
		self.packages = {}
		
		self.logger = None
		self.mount_p = "/"
		self.path_suffix_len = 0
		self.route_map = {}
		self.route_priority = []
		self._binds_server = [None] * 3
		self._binds_request = [None] * 4
		self._binds_when = [None] * 5		
				
	def set_mount_point (self, mount):	
		if not mount:
			self.mount_p = "/"
		elif mount [-1] != "/":
			self.mount_p = mount + "/"
		else:
			self.mount_p = mount
		self.path_suffix_len = len (self.mount_p) - 1
				
	def init (self, module, packagename = "app", mount = "/"):
		self.module = module	
		self.packagename = packagename
		self.set_mount_point (mount)
		
		if self.module:
			self.abspath = self.module.__file__
			if self.abspath [-3:] != ".py":
				self.abspath = self.abspath [:-1]
			self.update_file_info	()
		
	def __getitem__ (self, k):
		return self.route_map [k]
	
	def log (self, msg):
		self.logger.log (msg)
	
	def trace (self):
		self.logger.trace ()
		
	def reload_package (self):
		importer.reloader (self.module)
		self.update_file_info	()
	
	def is_reloadable (self):
		if self.module is None: return False
		stat = os.stat (self.abspath)
		return self.file_info != (stat.st_mtime, stat.st_size)		
			
	def update_file_info (self):
		stat = os.stat (self.abspath)
		self.file_info = (stat.st_mtime, stat.st_size)
	
	#----------------------------------------------
	# Decorators
	#----------------------------------------------
	def startup (self, f):
		self._binds_server [0] = f
	
	def onreload (self, f):
		self._binds_server [1] = f
		
	def shutdown (self, f):
		self._binds_server [2] = f
		
	def before_request (self, f):
		self._binds_request [0] = f
	
	def finish_request (self, f):
		self._binds_request [1] = f
	
	def failed_request (self, f):
		self._binds_request [2] = f
	
	def teardown_request (self, f):
		self._binds_request [3] = f
	
	def got_template (self, f):
		self._binds_when [0] = f
	
	def template_rendered (self, f):
		self._binds_when [1] = f
	
	#----------------------------------------------
	# Event Binding
	#----------------------------------------------
	def when_got_template (self, *args):
		self._binds_when [0] and self._binds_when [0] (*args)
	
	def when_template_rendered (self, *args):
		self._binds_when [1] and self._binds_when [1] (*args)
		
	def when_message_flashed (self, *args):
		self._binds_when [2] and self._binds_when [2] (*args)
	
	#----------------------------------------------
	# URL Building
	#----------------------------------------------		
	def url_for (self, thing, *args, **kargs):
		if thing.startswith ("/"):
			return self.route [:-1] + self.mount_p [:-1] + thing
	
		for func, name, fuvars, favars, numvars, str_rule, options in self.route_map.values ():			 
			if thing != name: continue								
			assert len (args) <= len (fuvars), "Too many params, this has only %d params(s)" % len (fuvars)						
			params = {}
			for i in range (len (args)):
				assert fuvars [i] not in kargs, "Collision detected on keyword param '%s'" % fuvars [i]
				params [fuvars [i]] = args [i]				
			
			for k, v in kargs.items ():
				params [k] = v
			
			url = str_rule
			if favars: #fancy [(name, type),...]. /fancy/<int:cid>/<cname>
				for n, t in favars:
					if n not in params:
						raise AssertionError ("Argument '%s' missing" % n)
					value = quote_plus (str (params [n]))
					if t == "string":
						value = value.replace ("+", "_")
					elif t == "path":
						value = value.replace ("%2F", "/")
					url = url.replace ("<%s%s>" % (t != "string" and t + ":" or "", n), value)
					del params [n]
			
			if params:
				url = url + "?" + "&".join (["%s=%s" % (k, quote_plus (str(v))) for k, v in params.items ()])
				
			return self.url_for (url)
	
	def build_url (self, thing, *args, **kargs):
		url = self.url_for (thing, *args, **kargs)
		if url:
			return url			
		
		for p in self.packages:
			url = self.packages [p].build_url (thing, *args, **kargs)
			if url:
				return url
				
	#----------------------------------------------
	# Routing
	#----------------------------------------------						
	def route (self, rule, **k):
		def decorator(f):
			self.add_route (rule, f, **k)
		return decorator
	
	def get_route_map (self):
		return self.route_map
	
	def set_route_map (self, route_map):
		self.route_map = route_map
	
	def mount (self, mount, module, partname = "app"):
		part = getattr (module, partname)
		part.init (module, partname, self.mount_p [:-1] + mount)
		self.packages [id (part)] = part
									
	def try_rule (self, path_info, rule, rulepack):
		f, n, l, a, c, s, options = rulepack
		
		arglist = rule.findall (path_info)
		if not arglist: 
			return None, None
		
		arglist = arglist [0]
		if type (arglist) is not tuple:
			arglist = (arglist,)
			
		kargs = {}
		for i in range(len(arglist)):
			an, at = a [i]
			if at == "int":
				kargs [an] = int (arglist [i])
			elif at == "float":
				kargs [an] = float (arglist [i])
			elif at == "path":
				kargs [an] = unquote_plus (arglist [i])
			else:		
				kargs [an] = unquote_plus (arglist [i]).replace ("_", " ")
		return f, kargs
	
	def add_route (self, rule, func, **options):					
		if not rule or rule [0] != "/":
			raise AssertionError ("Url rule should be starts with '/'")
		
		fspec = inspect.getargspec(func)
		if fspec [3]:
			defaults = {}
			argnames = fspec [0][(len (fspec [0]) - len (fspec [3])):]
			for i in range (len (fspec [3])):
				defaults [argnames [i]] = fspec [3][i]
			options ["defaults"] = defaults
		
		s = rule.find ("/<")
		if s == -1:	
			self.route_map [rule] = (func, func.__name__, func.__code__.co_varnames [1:func.__code__.co_argcount], None, func.__code__.co_argcount - 1, rule, options)			
		else:
			s_rule = rule
			rulenames = []
			for r, n in RX_RULE.findall (rule):
				if n.startswith ("int:"):
					rulenames.append ((n[4:], n[:3]))
					rule = rule.replace (r, "/([0-9]+)")
				elif n.startswith ("float:"):
					rulenames.append ((n[6:], n [:5]))
					rule = rule.replace (r, "/([.0-9]+)")
				elif n.startswith ("path:"):
					rulenames.append ((n[5:], n [:4]))
					rule = rule.replace (r, "/(.+)")	
				else:
					rulenames.append ((n, "string"))
					rule = rule.replace (r, "/([^/]+)")
			rule = "^" + rule + "$"			
			re_rule = re.compile (rule)				
			self.route_map [re_rule] = (func, func.__name__, func.__code__.co_varnames [1:func.__code__.co_argcount], tuple (rulenames), func.__code__.co_argcount - 1, s_rule, options)
			self.route_priority.append ((s, re_rule))
			self.route_priority.sort (key = lambda x: x [0], reverse = True)			
	
	def get_routed (self, method_pack):
		if not method_pack: 
			return
		temp = method_pack
		while 1:
			routed = temp [1]
			if type (routed) is not list:
				return routed
			temp = routed
					
	def route_search (self, path_info):
		if not path_info:
			return self.url_for ("/"), self.route_map ["/"]
		if path_info in self.route_map:
			return self.route_map [path_info][0], self.route_map [path_info]
		trydir = path_info + "/"
		if trydir in self.route_map:
			return self.url_for (trydir), self.route_map [trydir]
		raise KeyError
					
	def get_package_method (self, path_info, command, content_type, authorization, use_reloader = False):		
		if not (path_info.startswith (self.mount_p) or (path_info + "/").startswith (self.mount_p)):
			return None, None, None, None, None, 0
		
		path_info = path_info [self.path_suffix_len:]
		app, method, kargs, matchtype = self, None, {}, 0				
		# 1st, try find in self
		try:			
			method, current_rule = self.route_search (path_info)
			
		except KeyError:
			for priority, rule in self.route_priority:
				current_rule = self.route_map [rule]
				method, kargs = self.try_rule (path_info, rule, current_rule)
				if method: 
					match = (rule, current_rule)
					matchtype = 2
					options = current_rule [-1]
					break
		
		else:			
			match = path_info
			if type (method) is not function:
				# object move
				matchtype = -1
			else:	
				matchtype = 1
				options = current_rule [-1]
			
		# 2nd, try find in sub packages		
		if method is None and self.packages:
			for pid in list (self.packages.keys ()):
				package = self.packages [pid]
				subapp = getattr (package.module, package.packagename)
								
				if use_reloader and subapp.is_reloadable ():
					del self.packages [pid]
					args, its_packages = (package.mount_p, package.module, package.packagename), package.packages
					subapp.reload_package ()
					self.mount (*args)
					package.start (self.wasc, self.route, its_packages)
					subapp = getattr (package.module, package.packagename)
										
				app, method, kargs, options, match, matchtype = subapp.get_package_method (path_info, method, content_type, authorization, use_reloader)				
				if method:
					break
		
		if method is None:
			return None, None, None, None, None, 0
		
		if matchtype == -1: # 301 move
			return app, method, None, None, None, -1
			
		return app, [self._binds_request [0], method] + self._binds_request [1:4], kargs, options, match, matchtype
	
	#----------------------------------------------
	# Starting App
	#----------------------------------------------
	def cleanup (self):
		# initing app & packages		
		self._binds_server [2] and self._binds_server [2] (self.wasc)
		for p in list (self.packages.values ()):
			p.cleanup ()
			
	def start (self, wasc, route, packages = None):
		self.wasc = wasc
		if not route: 
			self.route = "/"
		elif not route.endswith ("/"):			
			self.route = route + "/"
		else:
			self.route = route
			
		if packages is None:
			# initing app & packages
			self._binds_server [0] and self._binds_server [0] (self.wasc)
			for p in list (self.packages.values ()):
				p.start (self.wasc, route)
				
		else:
			self._binds_server [1] and self._binds_server [1] (self.wasc)
			self.packages = packages




# -*- Mode: Python; tab-width: 4 -*-

import re
import string
import time

def concat (*args):
	return ''.join (args)	

def join (seq, field=' '):
	return field.join (seq)

def group (s):
	return '(' + s + ')'

short_days = ['sun','mon','tue','wed','thu','fri','sat']
long_days = ['sunday','monday','tuesday','wednesday','thursday','friday','saturday']

short_day_reg = group (join (short_days, '|'))
long_day_reg = group (join (long_days, '|'))

daymap = {}
for i in range(7):
	daymap[short_days[i]] = i
	daymap[long_days[i]] = i

hms_reg = join (3 * [group('[0-9][0-9]')], ':')

months = ['jan','feb','mar','apr','may','jun','jul','aug','sep','oct','nov','dec']

monmap = {}
for i in range(12):
	monmap[months[i]] = i+1

months_reg = group (join (months, '|'))

# From draft-ietf-http-v11-spec-07.txt/3.3.1
#       Sun, 06 Nov 1994 08:49:37 GMT  ; RFC 822, updated by RFC 1123
#       Sunday, 06-Nov-94 08:49:37 GMT ; RFC 850, obsoleted by RFC 1036
#       Sun Nov  6 08:49:37 1994       ; ANSI C's asctime() format

# rfc822 format
rfc822_date = join (
	[concat (short_day_reg,','),	# day
	 group('[0-9][0-9]?'),			# date
	 months_reg,					# month
	 group('[0-9]+'),				# year
	 hms_reg,						# hour minute second
	 'gmt'
	 ],
	' '
	)

rfc822_reg = re.compile (rfc822_date)

def unpack_rfc822 (m):
	g = m.group
	a = int
	return (
		a(g(4)),	   	# year
		monmap[g(3)],	# month
		a(g(2)),		# day
		a(g(5)),		# hour
		a(g(6)),		# minute
		a(g(7)),		# second
		0,
		0,
		0
		)

# rfc850 format
rfc850_date = join (
	[concat (long_day_reg,','),
	 join (
		 [group ('[0-9][0-9]?'),
		  months_reg,
		  group ('[0-9]+')
		  ],
		 '-'
		 ),
	 hms_reg,
	 'gmt'
	 ],
	' '
	)

rfc850_reg = re.compile (rfc850_date)
# they actually unpack the same way
def unpack_rfc850 (m):
	g = m.group
	a = int
	return (
		a(g(4)),	   	# year
		monmap[g(3)],	# month
		a(g(2)),		# day
		a(g(5)),		# hour
		a(g(6)),		# minute
		a(g(7)),		# second
		0,
		0,
		0
		)

# parsdate.parsedate	- ~700/sec.
# parse_http_date    	- ~1333/sec.

def build_http_date (when):
	return time.strftime ('%a, %d %b %Y %H:%M:%S GMT', time.gmtime(when))

def parse_http_date (d):
	d = d.lower ()
	tz = time.timezone
	m = rfc850_reg.match (d)
	if m and m.end() == len(d):
		retval = int (time.mktime (unpack_rfc850(m)) - tz)
	else:
		m = rfc822_reg.match (d)
		if m and m.end() == len(d):
			retval = int (time.mktime (unpack_rfc822(m)) - tz)
		else:
			return 0
	# Thanks to Craig Silverstein <csilvers@google.com> for pointing
	# out the DST discrepancy
	if time.daylight and time.localtime(retval)[-1] == 1: # DST correction
		retval = retval + (tz - time.altzone)
	return retval



from .secured_cookie_value import SecuredListValue
import time

class MessageBox (SecuredListValue):
	KEY = "MBOX"
	
	def __init__ (self, name, cookie, request, secret_key):
		SecuredListValue.__init__ (self, name, cookie, request, secret_key)
		self.mid = -1
	
	def serialize (self):
		self.data.append ((-1, time.time (), self.request.get_remote_addr ())) # add random string
		return SecuredListValue.serialize (self)
		
	def validate (self):	
		if not self.data:
			self.data = []
			return
			
		if self.data and self.data [0][0] != -1:
			return
			
		validator, self.data = self.data [0], self.data [1:]
		last_update, addr = validator [1:3]
		self.__source_verified = (addr == self.request.get_remote_addr ())
					
	def send (self, msg, category = "info", valid = 0, **extra):
		self.data is None and self.unserialize ()
		if self.data and self.mid == -1:
			self.mid = max ([n [0] for n in self.data])		
		self.mid += 1
		self.data.append ((self.mid, category, int (time.time ()), valid, msg, extra))
		self.dirty = True
		
	def remove (self, mid):
		self.data is None and self.unserialize ()
		index = 0
		found = False
		for n in self.data:
			if n [0] == mid:
				found = True
				break
			index += 1
		if found:
			self.data.pop (index)
		self.dirty = True
	
	def search (self, k, v = None):
		self.data is None and self.unserialize ()
		mids = []
		for notice in self.data:
			if v is None:
				if notice [1] == k:
					mids.append (notice [0])
			elif notice [5].get (k) == v:
				mids.append (notice [0])
		return mids
	
	def getv (self, k = None, v = None):
		if not self.__source_verified:
			self.data = []
			self.dirty = True
			return []
		return self.get (k, v)	
		
	def get (self, k = None, v = None):
		self.data is None and self.unserialize ()
		mids = []
		if k:
			mids = self.search (k, v)

		now = int (time.time ())
		messages = []
		not_expired = []
		
		for notice in self.data:
			how_old = now - notice [2]			
			if notice [3] and how_old > notice [3]:
				# expired, drop
				continue
				
			if mids and notice [0] not in mids:
				not_expired.append (notice)
				continue
				
			if notice [3]:
				not_expired.append (notice)			
							
			messages.append (notice)
		
		if len (self.data) != len (not_expired):
			self.data = not_expired
			self.dirty = True
			
		return messages
	
	def recal_expires (self, expires):						
		if self.data:
			return "never"
		return 0
	




from aquests.lib import attrdict

def Config (preset = False):
	cf = attrdict.AttrDict ()
	if preset:
		cf.max_post_body_size = 5 * 1024 * 1024
		cf.max_cache_size = 5 * 1024 * 1024
		cf.max_multipart_body_size = 20 * 1024 * 1024
		cf.max_upload_file_size = 20000000
	return cf
